import numpy as np
from pycub import pyCub
import open3d as o3d
import open3d.visualization.rendering as rendering


def draw_lines(client):
    """
    Draws lines from the head to the ball and from the head in the direction of the gaze
    """

    # Prepare line material for open3d
    mat = rendering.MaterialRecord()
    mat.shader = "unlitLine"
    mat.line_width = 10

    # get ball position
    ball_pos, _ = client.getBasePositionAndOrientation(client.free_objects[0])

    # Get head link position and orientation
    head_state = client.getLinkState(client.robot, 96, computeLinkVelocity=0, computeForwardKinematics=0)
    head_pos, head_ori = head_state[0], head_state[1]

    # The direction if a vector from the ball to the head
    head_ball_direction = np.array(ball_pos) - head_pos
    head_ball_len = np.linalg.norm(head_ball_direction)
    head_ball_direction /= head_ball_len # normalize

    # draw the line
    client.visualizer.vis.remove_geometry("ball_line")
    lb = o3d.geometry.LineSet(o3d.utility.Vector3dVector([head_pos, head_pos + head_ball_len * head_ball_direction]),
                              o3d.utility.Vector2iVector([[0, 1]]))
    lb.colors = o3d.utility.Vector3dVector([[1, 0, 0]])
    client.visualizer.vis.add_geometry("ball_line", geometry=lb, material=mat)

    # Direction of the gaze is the Z-axis rotated by the head orientation
    R_l = np.eye(4)
    R_l[:3, :3] = np.reshape(client.getMatrixFromQuaternion(head_ori), (3, 3))
    head_direction = np.matmul(R_l, [0, 0, 1, 1])[:3]  # Rotate Z-axis vector to point in direction of gripper
    head_direction /= np.linalg.norm(head_direction)  # normalize

    # Draw the line
    client.visualizer.vis.remove_geometry("head_line")
    lh = o3d.geometry.LineSet(o3d.utility.Vector3dVector([head_pos, head_pos + head_ball_len * head_direction]),
                              o3d.utility.Vector2iVector([[0, 1]]))
    lh.colors = o3d.utility.Vector3dVector([[0, 0, 1]])
    client.visualizer.vis.add_geometry("head_line", geometry=lh, material=mat)

    return head_direction, head_ball_direction


if __name__ == "__main__":
    client = pyCub(config="gaze.yaml")
    # look down and move arms from the view
    client.move_position("neck_pitch", -0.69, wait=False)
    client.move_position(["l_shoulder_pitch", "l_shoulder_roll", "r_shoulder_pitch", "r_shoulder_roll"],
                         [-1.5, 1.5, -1.5, 1.5], wait=False)
    while not client.motion_done():
        head_direction, head_ball_direction = draw_lines(client)
        client.update_simulation(None)

    # to get the mutual position of two vectors "up" vector is needed
    # can be computed as one of the orthogonal vectors to the vectors

    # null_space = scipy.linalg.null_space(head_direction.reshape(1, 3))
    # v = null_space[:, 1]

    # fixed here so scipy is not needed
    v = np.array([-6.30446773e-01,  6.00250287e-04,  7.76232250e-01])

    K = -1
    while client.is_alive():
        # apply force to the ball so it moves
        client.applyExternalForce(client.free_objects[0], -1, [0, K*1e-2, 0], [0, 0, 0], client.WORLD_FRAME)

        # draw lines
        head_direction, head_ball_direction = draw_lines(client)

        # get sign -> 1 when ball-head vector if "left" (from iCub point-of-view) from head vector
        sign = np.sign(np.dot(np.cross(head_direction, head_ball_direction), v))

        # angle between the vectors
        angle = sign*np.arctan2(np.linalg.norm(np.cross(head_direction, head_ball_direction)), np.dot(head_direction, head_ball_direction))

        # YOUR CODE HERE

        client.update_simulation()

